#Try to do things a bit like https://github.com/python/cpython/blob/master/Lib/collections/__init__.py
#Except instead of putting things here directly in __init__, we'll impor them so they are accessible by importing the module.
from CheKiPEUQ.InverseProblem import *
from CheKiPEUQ.plotting_functions import *
from CheKiPEUQ.mumpce import *
#from InverseProblem import parameter_estimation